#pragma once

#include <cstdlib>
#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/voxel_painter.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/sprite.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/color.hpp>

namespace BGTU::ComputerGraphicsLabWork::Utilities::Shapes {
	class rectangle_fill_iterable {
	private:
		RendererApi::PointI<2>::component_t x_start, y_start, x_end, y_end;

		constexpr rectangle_fill_iterable(
				RendererApi::PointI<2>::component_t x0,
				RendererApi::PointI<2>::component_t y0,
				RendererApi::PointI<2>::component_t x1,
				RendererApi::PointI<2>::component_t y1
		) noexcept: x_start{x0}, y_start{y0}, x_end{x1}, y_end{y1} {}

	public:
		static constexpr rectangle_fill_iterable from_wh(
				RendererApi::PointI<2>::component_t x,
				RendererApi::PointI<2>::component_t y,
				RendererApi::PointI<2>::component_t w,
				RendererApi::PointI<2>::component_t h
		) noexcept {
			return {x, y, x + w, y + h};
		}

		static constexpr rectangle_fill_iterable from_edges(
				RendererApi::PointI<2>::component_t left,
				RendererApi::PointI<2>::component_t top,
				RendererApi::PointI<2>::component_t right,
				RendererApi::PointI<2>::component_t bottom
		) noexcept {
			return {left, top, right, bottom};
		}

		class iterator {
		private:
			RendererApi::PointI<2>::component_t x_start, x_end, y_end;
			RendererApi::PointI<2>::component_t x, y;

			constexpr iterator(
					RendererApi::PointI<2>::component_t x0,
					RendererApi::PointI<2>::component_t y0,
					RendererApi::PointI<2>::component_t x1,
					RendererApi::PointI<2>::component_t y1
			) noexcept: x_start{x0}, x_end{x1}, y_end{y1}, x{x0}, y{y0} {}

			friend class rectangle_fill_iterable;

		public:
			constexpr iterator &operator++() noexcept {
				if (++this->x > this->x_end) {
					if (this->y < this->y_end) {
						this->y++;
					}
					this->x = x_start;
				}

				return *this;
			}

			constexpr iterator operator++(int) noexcept {
				iterator next = *this;
				++(*this);
				return next;
			}

			constexpr RendererApi::PointI<2> operator*() const noexcept {
				return {this->x, this->y};
			}

			constexpr bool operator==(std::nullptr_t) const noexcept {
				return this->y >= this->y_end;
			}

			constexpr bool operator!=(std::nullptr_t) const noexcept {
				return this->y < this->y_end;
			}
		};

		[[nodiscard]] constexpr iterator begin() const noexcept {
			return iterator{
					this->x_start, this->y_start,
					this->x_end, this->y_end
			};
		}

		[[nodiscard]] constexpr std::nullptr_t end() const noexcept {
			return nullptr;
		}
	};

	template<class receiver_t>
	void iterate_rectangle_fill_edges(
			RendererApi::PointI<2>::component_t left,
			RendererApi::PointI<2>::component_t top,
			RendererApi::PointI<2>::component_t right,
			RendererApi::PointI<2>::component_t bottom,
			receiver_t receiver
	) {
		for (RendererApi::PointI<2>::component_t x = left; x < right; x++) {
			receiver(true, x, top);
		}

		right--;
		bottom--;
		for (RendererApi::PointI<2>::component_t y = top+1; y < bottom; y++) {
			RendererApi::PointI<2>::component_t x = left;
			receiver(true, x, y);
			for (x++; x < right; x++) {
				receiver(false, x, y);
			}
			receiver(true, x, y);
		}

		for (RendererApi::PointI<2>::component_t x = left; x <= right; x++) {
			receiver(true, x, bottom);
		}
	}

	template<class receiver_t>
	void iterate_rectangle_fill_wh(
			RendererApi::PointI<2>::component_t x,
			RendererApi::PointI<2>::component_t y,
			RendererApi::PointI<2>::component_t width,
			RendererApi::PointI<2>::component_t height,
			receiver_t receiver
	) {
		return iterate_rectangle_fill_edges<receiver_t>(x, y, x + width, y + height, receiver);
	}

	template<class voxel_painter_t = RendererApi::VirtualVoxelPainter>
	void draw_rectangle_edges(
			voxel_painter_t *painter,
			RendererApi::PointI<2>::component_t left,
			RendererApi::PointI<2>::component_t top,
			RendererApi::PointI<2>::component_t right,
			RendererApi::PointI<2>::component_t bottom,
			double z,
			RendererApi::Color::Transparent outline
	) {
		iterate_rectangle_fill_edges(
				left, top, right, bottom,
				[=](bool is_edge, RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) { if (is_edge) painter->add_voxel(x, y, z, outline); }
		);
	}

	template<class voxel_painter_t = RendererApi::VirtualVoxelPainter>
	void draw_rectangle_wh(
			voxel_painter_t *painter,
			RendererApi::PointI<2>::component_t x,
			RendererApi::PointI<2>::component_t y,
			RendererApi::PointI<2>::component_t width,
			RendererApi::PointI<2>::component_t height,
			double z,
			RendererApi::Color::Transparent outline
	) {
		iterate_rectangle_fill_wh(
				x, y, width, height,
				[=](bool is_edge, RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) { if (is_edge) painter->add_voxel(x, y, z, outline); }
		);
	}

	template<class voxel_painter_t = RendererApi::VirtualVoxelPainter>
	void fill_rectangle_edges(
			voxel_painter_t *painter,
			RendererApi::PointI<2>::component_t left,
			RendererApi::PointI<2>::component_t top,
			RendererApi::PointI<2>::component_t right,
			RendererApi::PointI<2>::component_t bottom,
			double z,
			RendererApi::Color::Transparent outline,
			RendererApi::Color::Transparent fill
	) {
		iterate_rectangle_fill_edges(
				left, top, right, bottom,
				[=](bool is_edge, RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) { painter->add_voxel(x, y, z, is_edge ? outline : fill); }
		);
	}

	template<class voxel_painter_t = RendererApi::VirtualVoxelPainter>
	void fill_rectangle_wh(
			voxel_painter_t *painter,
			RendererApi::PointI<2>::component_t x,
			RendererApi::PointI<2>::component_t y,
			RendererApi::PointI<2>::component_t width,
			RendererApi::PointI<2>::component_t height,
			double z,
			RendererApi::Color::Transparent outline,
			RendererApi::Color::Transparent fill
	) {
		iterate_rectangle_fill_wh(
				x, y, width, height,
				[=](bool is_edge, RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) { painter->add_voxel(x, y, z, is_edge ? outline : fill); }
		);
	}

	template<class voxel_painter_t = RendererApi::VirtualVoxelPainter>
	void draw_rectangle_edges(
			voxel_painter_t *painter,
			RendererApi::PointI<2>::component_t left,
			RendererApi::PointI<2>::component_t top,
			RendererApi::PointI<2>::component_t right,
			RendererApi::PointI<2>::component_t bottom,
			double z,
			RendererApi::Color::Transparent outline,
			RendererApi::SpriteMetadata *owner
	) {
		iterate_rectangle_fill_edges(
				left, top, right, bottom,
				[=](bool is_edge, RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) { if (is_edge) painter->add_voxel(x, y, z, outline, owner); }
		);
	}

	template<class voxel_painter_t = RendererApi::VirtualVoxelPainter>
	void draw_rectangle_wh(
			voxel_painter_t *painter,
			RendererApi::PointI<2>::component_t x,
			RendererApi::PointI<2>::component_t y,
			RendererApi::PointI<2>::component_t width,
			RendererApi::PointI<2>::component_t height,
			double z,
			RendererApi::Color::Transparent outline,
			RendererApi::SpriteMetadata *owner
	) {
		iterate_rectangle_fill_wh(
				x, y, width, height,
				[=](bool is_edge, RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) { if (is_edge) painter->add_voxel(x, y, z, outline, owner); }
		);
	}

	template<class voxel_painter_t = RendererApi::VirtualVoxelPainter>
	void fill_rectangle_edges(
			voxel_painter_t *painter,
			RendererApi::PointI<2>::component_t left,
			RendererApi::PointI<2>::component_t top,
			RendererApi::PointI<2>::component_t right,
			RendererApi::PointI<2>::component_t bottom,
			double z,
			RendererApi::Color::Transparent outline,
			RendererApi::Color::Transparent fill,
			RendererApi::SpriteMetadata *owner
	) {
		iterate_rectangle_fill_edges(
				left, top, right, bottom,
				[=](bool is_edge, RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) { painter->add_voxel(x, y, z, (is_edge ? outline : fill), owner); }
		);
	}

	template<class voxel_painter_t = RendererApi::VirtualVoxelPainter>
	void fill_rectangle_wh(
			voxel_painter_t *painter,
			RendererApi::PointI<2>::component_t x,
			RendererApi::PointI<2>::component_t y,
			RendererApi::PointI<2>::component_t width,
			RendererApi::PointI<2>::component_t height,
			double z,
			RendererApi::Color::Transparent outline,
			RendererApi::Color::Transparent fill,
			RendererApi::SpriteMetadata *owner
	) {
		iterate_rectangle_fill_wh(
				x, y, width, height,
				[=](bool is_edge, RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) { painter->add_voxel(x, y, z, (is_edge ? outline : fill), owner); }
		);
	}
}